/*
htop
(C) 2004 Hisham H. Muhammad
Released under the GNU GPL, see the COPYING file
in the source distribution for its full text.
*/

#include "FunctionBar.h"
#include "CRT.h"

#include "debug.h"

#include <string.h>
#include <stdlib.h>
#include <stdbool.h>
#include <curses.h>

/*{

typedef struct FunctionBar_ {
   int size;
   char** functions;
   char** keys;
   int* events;
} FunctionBar;

}*/

FunctionBar* FunctionBar_new(int size, char** functions, char** keys, int* events) {
   FunctionBar* this = malloc(sizeof(FunctionBar));
   this->size = size;
   this->functions = functions;
   this->keys = keys;
   this->events = events;
   return this;
}

void FunctionBar_delete(FunctionBar* this) {
   // These are always static data, RIGHT? >;)
   // free(this->functions);
   // free(this->keys);
   free(this);
}

void FunctionBar_draw(FunctionBar* this, char* buffer) {
   attron(COLOR_PAIR(FUNCTIONBAR_PAIR));
   mvhline(LINES-1, 0, ' ', COLS);
   attroff(COLOR_PAIR(FUNCTIONBAR_PAIR));
   int x = 0;
   for (int i = 0; i < this->size; i++) {
      attron(COLOR_PAIR(WHITE_PAIR));
      mvaddstr(LINES-1, x, this->keys[i]);
      attroff(COLOR_PAIR(WHITE_PAIR));
      x += strlen(this->keys[i]);
      attron(COLOR_PAIR(FUNCTIONBAR_PAIR));
      mvaddstr(LINES-1, x, this->functions[i]);
      attroff(COLOR_PAIR(FUNCTIONBAR_PAIR));
      x += strlen(this->functions[i]);
   }
   if (buffer != NULL) {
      attron(COLOR_PAIR(FUNCTIONBAR_PAIR));
      mvaddstr(LINES-1, x, buffer);
      attroff(COLOR_PAIR(FUNCTIONBAR_PAIR));
   }
}

int FunctionBar_synthesizeEvent(FunctionBar* this, int pos) {
   int x = 0;
   for (int i = 0; i < this->size; i++) {
      x += strlen(this->keys[i]);
      x += strlen(this->functions[i]);
      if (pos < x) {
         return this->events[i];
      }
   }
   return ERR;
}
