/*
htop
(C) 2004 Hisham H. Muhammad
Released under the GNU GPL, see the COPYING file
in the source distribution for its full text.
*/

#include "ProcessList.h"
#include "CRT.h"
#include "ListBox.h"
#include "UsersTable.h"
#include "Signal.h"
#include "RichString.h"
#include "ProcessFilter.h"
#include "Settings.h"
#include "ScreenManager.h"
#include "FunctionBar.h"
#include "ListItem.h"

#include "config.h"
#include "debug.h"

#include <unistd.h>
#include <math.h>
#include <sys/param.h>
#include <ctype.h>
#include <stdbool.h>

int usleep(int usec);

//#link m

#define INCSEARCH_MAX 40

/* private property */
char htop_barCharacters[] = "|#*@$%&";

void printVersionFlag() {
   clear();
   printf("htop " VERSION " - (C) 2004 Hisham Muhammad.\n");
   printf("Released under the GNU GPL.\n\n");
   exit(0);
}

void printHelpFlag() {
   clear();
   printf("htop " VERSION " - (C) 2004 Hisham Muhammad.\n");
   printf("Released under the GNU GPL.\n\n");
   printf("Press F1 inside htop for online help.\n");
   printf("See the man page for full info.\n\n");
   exit(0);
}

void showHelp() {
   clear();
   attron(A_BOLD);
   mvaddstr(0, 0, "htop " VERSION " - (C) 2004 Hisham Muhammad.");
   attroff(A_BOLD);
   mvaddstr(1, 0, "Released under the GNU GPL.");
   mvaddstr(3, 0, "CPU usage bar: [");
   attron(COLOR_PAIR(BLUE_PAIR));
   addstr("low-priority");
   attroff(COLOR_PAIR(BLUE_PAIR));
   addstr("/");
   attron(COLOR_PAIR(GREEN_PAIR));
   addstr("normal");
   attroff(COLOR_PAIR(GREEN_PAIR));
   addstr("/");
   attron(COLOR_PAIR(RED_PAIR));
   addstr("kernel");
   attroff(COLOR_PAIR(RED_PAIR));
   attron(A_BOLD | COLOR_PAIR(BLACK_PAIR));
   addstr("      used%");
   attroff(A_BOLD | COLOR_PAIR(BLACK_PAIR));
   addstr("]");
   mvaddstr(4, 0, "Memory bar:    [");
   attron(COLOR_PAIR(GREEN_PAIR));
   addstr("used");
   attroff(COLOR_PAIR(GREEN_PAIR));
   addstr("/");
   attron(COLOR_PAIR(BLUE_PAIR));
   addstr("buffers");
   attroff(COLOR_PAIR(BLUE_PAIR));
   addstr("/");
   attron(COLOR_PAIR(BROWN_PAIR));
   addstr("cache");
   attroff(COLOR_PAIR(BROWN_PAIR));
   attron(A_BOLD | COLOR_PAIR(BLACK_PAIR));
   addstr("         used/total");
   attroff(A_BOLD | COLOR_PAIR(BLACK_PAIR));
   addstr("]");
   mvaddstr(5, 0, "Swap bar:      [");
   attron(COLOR_PAIR(RED_PAIR));
   addstr("used");
   attroff(COLOR_PAIR(RED_PAIR));
   attron(A_BOLD | COLOR_PAIR(BLACK_PAIR));
   addstr("                       used/total");
   attroff(A_BOLD | COLOR_PAIR(BLACK_PAIR));
   addstr("]");
   mvaddstr(6, 0, "CPU(0) is the average of all CPUs; CPU(1) is the first CPU, etc.");

   attron(A_BOLD);
   mvaddstr(8, 0, "Summary of commands -- see the man page for full info");
   attroff(A_BOLD);
   mvaddstr(9, 0, " Arrow keys - scroll process list");
   mvaddstr(10, 0, " / - incremental process search: type process name");
   mvaddstr(11, 0, " Numbers - incremental PID search: type process ID");
   mvaddstr(12, 0, " [ - decrease process priority");
   mvaddstr(13, 0, " ] - increase process priority (superuser only)");
   mvaddstr(14, 0, " k - kill process(es)");
   mvaddstr(15,0, " P - sort by CPU%                Enter - refresh");
   mvaddstr(16,0, " M - sort by MEM%                F - cursor follows process");
   mvaddstr(17,0, " C - configure columns           S - setup meters");
   mvaddstr(18,0, " h - shows this help screen      q - quit");
   attron(A_BOLD);
   mvaddstr(20,0, "Press any key to return.");
   attroff(A_BOLD);
   refresh();
   CRT_readKey();
   clear();
}

void showColumnConfig(ProcessFilter* pf) {

   int i;
   int startSelected = 0;
   int startAvailable = 0;
   int currRow = 0;
   int currCol = 0;
   bool configure = true;
   bool save = false;

   ProcessField avail[LAST_PROCESSFIELD + 1] = { 0 };
   ProcessField select[LAST_PROCESSFIELD + 1] = { 0 };
   ProcessField original[LAST_PROCESSFIELD + 1] = { 0 };
   int countAvail = 0;
   int countSelect = 0;
   
   for(i = 0; i < LAST_PROCESSFIELD && pf->fields[i] != LAST_PROCESSFIELD; i++) {
      select[i] = pf->fields[i];
      original[i] = pf->fields[i];
      countSelect++;
   }
   select[countSelect] = LAST_PROCESSFIELD;
   original[i] = pf->fields[i];
   for(i = 0; i < LAST_PROCESSFIELD; i++) {
      bool found = false;
      for(int j = 0; j < LAST_PROCESSFIELD && pf->fields[j] != LAST_PROCESSFIELD; j++) 
         if(i == pf->fields[j]) found = true;
      if(!found) {
         avail[countAvail] = i;
         countAvail++;
      }
   }
   avail[countAvail] = LAST_PROCESSFIELD;

   clear();
   mvaddstr(0, 0, "Column configuration");
   attron(COLOR_PAIR(BLUE_PAIR));
   mvaddstr(4, 1, "Selected Columns");
   attroff(COLOR_PAIR(BLUE_PAIR));
   attron(COLOR_PAIR(GREEN_PAIR));
   mvaddstr(4, (COLS / 2) + 1, "Available Columns");
   attroff(COLOR_PAIR(GREEN_PAIR));
   char* functions[5] = { "Move Up", "Move Down", "Move <->", "Apply ", "Cancel" };
   char* keys[5] = { "- ", "+ ", "Enter", "w ", "Esc" };
   int events[5] = { '-', '+', 13, 'w', 27 };
   FunctionBar* fuBar = FunctionBar_new(5, functions, keys, events);
   FunctionBar_draw(fuBar, NULL);
   
   while(configure) {

      for(i = 0; i < LAST_PROCESSFIELD; i++) 
         pf->fields[i] = select[i];
      pf->fields[LAST_PROCESSFIELD] = LAST_PROCESSFIELD;
         
      for(i = 0; i < LAST_PROCESSFIELD; i++) {
         int field = select[i + startSelected];
         if(field == LAST_PROCESSFIELD) break;
         if(i == (LINES  - 8)) break;
         mvhline(5 + i, 1, ' ', COLS / 2);
         mvaddstr(5 + i, 1, processFieldNames[field]);
      }
      for (; i < LINES - 8; i++)
         mvhline(5 + i, 1, ' ', COLS / 2);

      RichString str = RichString_new();
      ProcessFilter_display((Object*)pf, &str);
      if (str.len > 0) {
         int attr = A_REVERSE | COLOR_PAIR(GREEN_PAIR);
         attron(attr);
         RichString_applyAttr(&str, attr);
         move(2, 0);
         hline(' ', 512);
         mvaddchstr(2, 0, str.chstr);
         attroff(attr);
      }

      for(i = 0; i < LAST_PROCESSFIELD; i++) {
         int field = avail[i + startAvailable];
         if(field == LAST_PROCESSFIELD) break;
         if(i == (LINES - 8)) break;
         mvhline(5 + i, (COLS / 2) + 1, ' ', COLS / 2);
         mvaddstr(5 + i, (COLS / 2) + 1, processFieldNames[field]);
      }
      for (; i < LINES - 8; i++)
         mvhline(5 + i, (COLS / 2) + 1, ' ', COLS / 2);
      mvchgat(5 + currRow, (currCol) ? (COLS / 2) + 1 : 1, (COLS / 2) - 2,
         A_REVERSE, FOCUS_HIGHLIGHT_PAIR, NULL);

      refresh();

      int *numEntries = (currCol) ? &countAvail : &countSelect;
      int *notEntries = (currCol) ? &countSelect : &countAvail;
      int *start = (currCol) ? &startAvailable : &startSelected;
      int pos = currRow + *start;
      
      int c = getch();

      mvchgat(5 + currRow, (currCol) ? (COLS / 2) + 1 : 1, (COLS / 2) - 2,
         A_NORMAL, 0, NULL);

      switch(c) {
      case KEY_DOWN:
         if(currRow + *start == *numEntries - 1) break;
         if(currRow < LINES - 9) currRow++;
         else {
            if((*numEntries - *start) > (LINES - 8))
               (*start)++;
         }
         break;

      case KEY_NPAGE:
         // TODO: quick and dirty hack. Better improve.
         for (int i = 0; i < LINES - 9; i++) {
            if(currRow + *start == *numEntries - 1) break;
               if(currRow < LINES - 9) currRow++;
               else {
                  if((*numEntries - *start) > (LINES - 8))
                     (*start)++;
               }
         }
         break;

      case KEY_PPAGE:
         // TODO: quick and dirty hack. Better improve.
         for (int i = 0; i < LINES - 9; i++) {
            if(currRow > 0) currRow--;
            else {
               if(*start > 0)
                  (*start)--;
            }
         }
         break;

      case KEY_UP:
         if(currRow > 0) currRow--;
         else {
            if(*start > 0)
               (*start)--;
         }
         break;

      case KEY_LEFT:
         currCol = 0;
         if(currRow > *notEntries - 1) currRow = *notEntries - 1;
         break;

      case KEY_RIGHT:
         if(countAvail == 0) break;
         currCol = 1;
         if(currRow > *notEntries - 1) currRow = *notEntries - 1;
         break;

      case '}':
      case ']':
      case '+':
      case '.':
      case '=': {
         if(currRow  + *start == *numEntries - 1) break;
         ProcessField *array = (currCol) ? avail : select;
         ProcessField inv = array[pos];
         array[pos] = array[pos + 1];
         array[pos + 1] = inv;
         if(currRow < LINES - 9) currRow++; //From Key Down
         else {
            if((*numEntries - *start) > (LINES - 8))
               (*start)++;
         }
         break;
      }

      case '{':
      case '[':
      case '_':
      case ',':
      case '-': {
         if(currRow + *start == 0) break;
         ProcessField *array = (currCol) ? avail : select;
         ProcessField inv = array[pos];
         array[pos] = array[pos - 1];
         array[pos - 1] = inv;
         if(currRow > 0) currRow--; //From Key up
         else {
            if(*start > 0)
               (*start)--;
         }
         break;
      }

      case 0x0a:
      case 0x0d:
      case KEY_ENTER:
         if(*numEntries == 0) break;
         if(!currCol && *numEntries == 1) break;
         ProcessField *array = (currCol) ? avail : select;
         ProcessField *notarray = (currCol) ? select : avail;
         for(i = *notEntries + 2; i >=1; i--) {
	    notarray[i] = notarray[i-1];
         }
         notarray[0] = array[pos];
         (*notEntries)++;

         for(i = pos; pos < LAST_PROCESSFIELD; i++) {
            if(array[i] == LAST_PROCESSFIELD) break;
            array[i] = array[i + 1];
         }
         (*numEntries)--;
         array[*numEntries] = LAST_PROCESSFIELD;
         if(*start > 0) (*start)--;
         else
            if(pos > *numEntries - 1) currRow--;
	 
         currCol = currCol == 0 ? 1 : 0;
	 currRow = 0;
	 
         if(*numEntries == 0) {
            currCol = 0;
            currRow = 0;
         }
         break;

      case 27:
      case 'q':
         configure = false;
         break;

      case 'w':
         save = true;
         configure = false;
         break;

      default:
         break;
      }
   }

   if(save) {
      for(i = 0; i < LAST_PROCESSFIELD && select[i] != LAST_PROCESSFIELD; i++)
         pf->fields[i] = select[i];
      pf->fields[countSelect] = LAST_PROCESSFIELD;
   }
   else {
      for(i = 0; i < LAST_PROCESSFIELD && original[i] != LAST_PROCESSFIELD; i++)
         pf->fields[i] = original[i];
      pf->fields[countSelect] = LAST_PROCESSFIELD;
   }
   FunctionBar_delete(fuBar);

   clear();

}

void Header_draw(Settings* settings) {
   int headerHeight = settings->headerHeight;
   int headerPad = settings->headerPad;
   TypedVector* leftMeters = settings->leftMeters;
   TypedVector* rightMeters = settings->rightMeters;
   
   for (int y = 0; y < headerHeight; y++) {
      mvhline(y, 0, ' ', COLS);
   }
   for (int y = (headerPad / 2), i = 0; i < TypedVector_size(leftMeters); i++) {
      Meter* meter = (Meter*) TypedVector_get(leftMeters, i);
      meter->draw(meter, headerPad, y, COLS/2-(headerPad*2-1)-1);
      y += meter->h;
   }
   for (int y = (headerPad / 2), i = 0; i < TypedVector_size(rightMeters); i++) {
      Meter* meter = (Meter*) TypedVector_get(rightMeters, i);
      meter->draw(meter, COLS/2+headerPad, y, COLS/2-(headerPad*2-1)-1);
      y += meter->h;
   }
}

HandlerResult SignalListBox_eventHandler(ListBox* this, int ch) {
   int size = ListBox_getSize(this);
   if (isdigit(ch)) {
      int signal = ch-48 + this->state.i;
      for (int i = 0; i < size; i++)
         if (((Signal*) ListBox_get(this, i))->number == signal) {
            ListBox_setSelected(this, i);
            break;
         }
      this->state.i = signal * 10;
      if (this->state.i > 100)
         this->state.i = 0;
      return HANDLED;
   } else {
      this->state.i = 0;
   }
   if (ch == 13) {
      return BREAK_LOOP;
   }
   return IGNORED;
}

/* private */
typedef struct MetersState_ {
   Settings* settings;
   TypedVector* meters;
   ScreenManager* scr;
} MetersState;

HandlerResult MetersListBox_eventHandler(ListBox* this, int ch) {
   
   // TODO: proper attributes once I have decent inheritance
   Settings* settings = ((MetersState*)this->state.ptr)->settings;
   TypedVector* meters = ((MetersState*)this->state.ptr)->meters;
   ScreenManager* scr = ((MetersState*)this->state.ptr)->scr;
   
   int selected = ListBox_getSelectedIndex(this);
   HandlerResult result = IGNORED;

   switch(ch) {
      case KEY_F(4):
      case 't':
      {
         Meter* meter = (Meter*) TypedVector_get(meters, selected);
         MeterMode mode = ++(meter->mode);
         if (mode == LAST_METERMODE)
            mode = 1; // skip mode 0, "unset"
         Meter_setMode(meter, mode);
         ListBox_set(this, selected, (Object*) Meter_toListItem(meter));
         result = HANDLED;
         break;
      }
      case KEY_F(7):
      case '[':
      case '-':
      {
         TypedVector_moveUp(meters, selected);
         ListBox_moveSelectedUp(this);
         result = HANDLED;
         break;
      }
      case KEY_F(8):
      case ']':
      case '+':
      {
         TypedVector_moveDown(meters, selected);
         ListBox_moveSelectedDown(this);
         result = HANDLED;
         break;
      }
      case KEY_F(9):
      case KEY_DC:
      {
         if (selected < TypedVector_size(meters)) {
            TypedVector_remove(meters, selected);
            ListBox_remove(this, selected);
         }
         result = HANDLED;
         break;
      }
   }
   if (result == HANDLED) {
      Settings_calculateHeaderHeight(settings);
      Header_draw(settings);
      ScreenManager_resize(scr, scr->x1, settings->headerHeight, scr->x2, scr->y2);
   }
   return result;
}

ListBox* MetersListBox_new(Settings* settings, char* header, TypedVector* meters, ScreenManager* scr) {
   ListBox* lb = ListBox_new(1, 1, 1, 1, LISTITEM_CLASS, true);
   
   // TODO: proper attributes once I have decent inheritance
   lb->state.ptr = malloc(sizeof(MetersState));
   lb->freeStatePointer = true;
   ((MetersState*)lb->state.ptr)->settings = settings;
   ((MetersState*)lb->state.ptr)->meters = meters;
   ((MetersState*)lb->state.ptr)->scr = scr;

   lb->eventHandler = MetersListBox_eventHandler;
   ListBox_setHeader(lb, RichString_quickString(header));
   for (int i = 0; i < TypedVector_size(meters); i++) {
      Meter* meter = (Meter*) TypedVector_get(meters, i);
      ListBox_add(lb, (Object*) Meter_toListItem(meter));
   }
   return lb;
}

/* private */
typedef struct AvailableMetersState_ {
   Settings* settings;
   ListBox* leftBox;
   ListBox* rightBox;
   ScreenManager* scr;
} AvailableMetersState;

HandlerResult AvailableMetersListBox_eventHandler(ListBox* this, int ch) {
   
   // TODO: proper attributes once I have decent inheritance
   Settings* settings = ((AvailableMetersState*)this->state.ptr)->settings;
   ListBox* leftBox = ((AvailableMetersState*)this->state.ptr)->leftBox;
   ListBox* rightBox = ((AvailableMetersState*)this->state.ptr)->rightBox;
   ScreenManager* scr = ((AvailableMetersState*)this->state.ptr)->scr;
   
   ListItem* selected = (ListItem*) ListBox_getSelected(this);
   char* name = selected->text;
   HandlerResult result = IGNORED;

   switch(ch) {
      case KEY_F(5):
      case 'l':
      case 'L':
      {
         TypedVector* meters = settings->leftMeters;
         Settings_createMeter(settings, name, meters);
         Meter* meter = (Meter*) TypedVector_get(meters, TypedVector_size(meters) - 1);
         ListBox_add(leftBox, (Object*) Meter_toListItem(meter));
         result = HANDLED;
         break;
      }
      case KEY_F(6):
      case 'r':
      case 'R':
      {
         TypedVector* meters = settings->rightMeters;
         Settings_createMeter(settings, name, meters);
         Meter* meter = (Meter*) TypedVector_get(meters, TypedVector_size(meters) - 1);
         ListBox_add(rightBox, (Object*) Meter_toListItem(meter));
         result = HANDLED;
         break;
      }
   }
   if (result == HANDLED) {
      Settings_calculateHeaderHeight(settings);
      Header_draw(settings);
      ScreenManager_resize(scr, scr->x1, settings->headerHeight, scr->x2, scr->y2);
   }
   return result;
}

ListBox* AvailableMetersListBox_new(Settings* settings, ListBox* leftMeters, ListBox* rightMeters, ScreenManager* scr) {
   
   ListBox* lb = ListBox_new(1, 1, 1, 1, LISTITEM_CLASS, true);
   
   // TODO: proper attributes once I have decent inheritance
   lb->state.ptr = malloc(sizeof(AvailableMetersState));
   lb->freeStatePointer = true;
   ((AvailableMetersState*)lb->state.ptr)->settings = settings;
   ((AvailableMetersState*)lb->state.ptr)->leftBox = leftMeters;
   ((AvailableMetersState*)lb->state.ptr)->rightBox = rightMeters;
   ((AvailableMetersState*)lb->state.ptr)->scr = scr;
   
   lb->eventHandler = AvailableMetersListBox_eventHandler;
   ListBox_setHeader(lb, RichString_quickString("Available"));
   ListBox_add(lb, (Object*) ListItem_new(String_copy("Swap")));
   ListBox_add(lb, (Object*) ListItem_new(String_copy("Memory")));
   ListBox_add(lb, (Object*) ListItem_new(String_copy("Clock")));
   ListBox_add(lb, (Object*) ListItem_new(String_copy("Load")));
   ListBox_add(lb, (Object*) ListItem_new(String_copy("LoadAverage")));
   ListBox_add(lb, (Object*) ListItem_new(String_copy("Uptime")));
   ListBox_add(lb, (Object*) ListItem_new(String_copy("Tasks")));
   for (int i = 0; i <= settings->pl->processorCount; i++) {
      char buffer[50];
      sprintf(buffer, "CPU(%d)", i);
      ListBox_add(lb, (Object*) ListItem_new(String_copy(buffer)));
   }
   return lb;
   
}

void Setup_run(Settings* settings, int headerHeight) {

   char* fuFunctions[10] = {"      ", "      ", "      ", "Type  ", "Add L ",
       "Add R ", "MoveUp", "MoveDn", "Remove", "Done  "};
   char* fuKeys[10] = {" 1", " 2", " 3", " 4", " 5", " 6", " 7", " 8", " 9", "10"};
   int fuEvents[10] = {KEY_F(1), KEY_F(2), KEY_F(3), KEY_F(4), KEY_F(5),
                       KEY_F(6), KEY_F(7), KEY_F(8), KEY_F(9), KEY_F(10)};
   FunctionBar* fuBar = FunctionBar_new(10, fuFunctions, fuKeys, fuEvents);

   ScreenManager* scr = ScreenManager_new(0, headerHeight, 0, -1, HORIZONTAL);
   
   ListBox* lbCategories = ListBox_new(1, 1, 1, 1, LISTITEM_CLASS, true);
   ListBox_setHeader(lbCategories, RichString_quickString("Setup:"));
   ListBox_add(lbCategories, (Object*) ListItem_new(String_copy("Meters")));

   ListBox* lbLeftMeters = MetersListBox_new(settings, "Left", settings->leftMeters, scr);
   ListBox* lbRightMeters = MetersListBox_new(settings, "Right", settings->rightMeters, scr);
   ListBox* lbAvailableMeters = AvailableMetersListBox_new(settings, lbLeftMeters, lbRightMeters, scr);
   
   ScreenManager_add(scr, lbCategories, 15);
   ScreenManager_add(scr, lbLeftMeters, 20);
   ScreenManager_add(scr, lbRightMeters, 20);
   ScreenManager_add(scr, lbAvailableMeters, -1);
   ScreenManager_setFunctionBar(scr, fuBar);
   ListBox* lbFocus;
   int ch;
   ScreenManager_run(scr, &lbFocus, &ch);
   ScreenManager_delete(scr);
   ListBox_delete((Object*) lbCategories);
   ListBox_delete((Object*) lbLeftMeters);
   ListBox_delete((Object*) lbRightMeters);
   ListBox_delete((Object*) lbAvailableMeters);
   FunctionBar_delete(fuBar);
}

int main(int argc, char** argv) {

   if (argc > 0) {
      if (String_eq(argv[1], "--help")) {
         printHelpFlag();
      } else if (String_eq(argv[1], "--version")) {
         printVersionFlag();
      }
   }

   ListBox* lb;
   int quit = 0;
   int refreshTimeout = 0;
   int resetRefreshTimeout = 5;
   bool doRefresh = true;
   ProcessFilter* pf;
   Settings* settings;
   
   TypedVector* leftMeters = TypedVector_new(METER_CLASS, true);
   TypedVector* rightMeters = TypedVector_new(METER_CLASS, true);

   Signal** signals = NULL;
   ListBox* lbk = NULL;

   char incSearchBuffer[INCSEARCH_MAX];
   int incSearchIndex = 0;
   incSearchBuffer[0] = 0;
   bool incSearchMode = false;

   ProcessList* pl = NULL;
   UsersTable* ut = UsersTable_new();
   pf = ProcessFilter_new();
   CRT_init();
   pl = ProcessList_new(pf, ut);
   ProcessList_scan(pl);
   
   settings = Settings_new(pl, leftMeters, rightMeters);
   int headerHeight = Settings_calculateHeaderHeight(settings);
   
   lb = ListBox_new(0, headerHeight, COLS, LINES - headerHeight - 2, PROCESS_CLASS, false);
   RichString headerString = RichString_new();
   ((Object*)pf)->display((Object*)pf, &headerString);

   ListBox_setHeader(lb, headerString);
   
   char* searchFunctions[3] = {"Next  ", "Exit  ", " Search: "};
   char* searchKeys[3] = {"F3", "Esc", "  "};
   int searchEvents[3] = {KEY_F(3), 27, ERR};
   FunctionBar* searchBar = FunctionBar_new(3, searchFunctions, searchKeys, searchEvents);
   
   char* defaultFunctions[10] = {"Help  ", "Setup ", "Search", "Invert", "Sort<-",
       "Sort->", "Nice -", "Nice +", "Kill  ", "Quit  "};
   char* defaultKeys[10] = {" 1", " 2", " 3", " 4", " 5", " 6", " 7", " 8", " 9", "10"};
   int defaultEvents[10] = {KEY_F(1), KEY_F(2), KEY_F(3), KEY_F(4), KEY_F(5),
                            KEY_F(6), KEY_F(7), KEY_F(8), KEY_F(9), KEY_F(10)};
   FunctionBar* defaultBar = FunctionBar_new(10, defaultFunctions, defaultKeys, defaultEvents);
   
   FunctionBar_draw(defaultBar, NULL);
   
   int acc = 0;
   bool follow = false;
   while (!quit) {
      if (doRefresh) {
         incSearchIndex = 0;
         incSearchBuffer[0] = 0;
         int currPos = ListBox_getSelectedIndex(lb);
         int currPid = 0;
         int currScrollV = lb->scrollV;
         if (follow)
            currPid = ProcessList_get(pl, currPos)->pid;
         ProcessList_scan(pl);
         if (refreshTimeout == 0) {
            ProcessList_sort(pl);
            refreshTimeout = 1;
         }
         ListBox_prune(lb);
         int size = ProcessList_size(pl);
         for (int i = 0; i < size; i++) {
            Process* p = ProcessList_get(pl, i);
            ListBox_set(lb, i, (Object*)p);
            if ((!follow && i == currPos) || (follow && p->pid == currPid)) {
               ListBox_setSelected(lb, i);
               lb->scrollV = currScrollV;
            }
         }
      }
      doRefresh = true;
      
      Header_draw(settings);

      ListBox_draw(lb, true);
      int ch = getch();
      if (incSearchMode) {
         doRefresh = false;
         if (ch == KEY_F(3)) {
            int here = ListBox_getSelectedIndex(lb);
            int size = ProcessList_size(pl);
            int i = here+1;
            while (i != here) {
               if (i == size)
                  i = 0;
               Process* p = ProcessList_get(pl, i);
               if (String_contains_i(p->comm, incSearchBuffer)) {
                  ListBox_setSelected(lb, i);
                  break;
               }
               i++;
            }
         } else if (isalnum(ch) && (incSearchIndex < INCSEARCH_MAX)) {
            incSearchBuffer[incSearchIndex] = ch;
            incSearchIndex++;
            incSearchBuffer[incSearchIndex] = 0;
            FunctionBar_draw(searchBar, incSearchBuffer);
            for (int i = 0; i < ProcessList_size(pl); i++) {
               Process* p = ProcessList_get(pl, i);
               if (String_contains_i(p->comm, incSearchBuffer)) {
                  ListBox_setSelected(lb, i);
                  break;
               }
            }
         } else if ((ch == KEY_BACKSPACE) && (incSearchIndex > 0)) {
            incSearchIndex--;
            incSearchBuffer[incSearchIndex] = 0;
               FunctionBar_draw(searchBar, incSearchBuffer);
          } else if (ch != ERR) {
            incSearchMode = false;
            incSearchIndex = 0;
            incSearchBuffer[0] = 0;
            FunctionBar_draw(defaultBar, NULL);
         }
         continue;
      }
      if (isdigit(ch)) {
         int pid = ch-48 + acc;
         for (int i = 0; i < ProcessList_size(pl) && ((Process*) ListBox_getSelected(lb))->pid != pid; i++)
            ListBox_setSelected(lb, i);
         acc = pid * 10;
         if (acc > 100000)
            acc = 0;
         continue;
      } else {
         acc = 0;
      }

      if (ch == KEY_MOUSE) {
         MEVENT mevent;
         int ok = getmouse(&mevent);
         if (ok == OK) {
            if (mevent.y >= lb->y + 1 && mevent.y < LINES - 1) {
               ListBox_setSelected(lb, mevent.y - lb->y + lb->scrollV - 1);
               doRefresh = false;
               refreshTimeout = resetRefreshTimeout;
               follow = true;
               continue;
            } if (mevent.y == LINES - 1) {
               FunctionBar* bar;
               if (incSearchMode) bar = searchBar;
               else bar = defaultBar;
               ch = FunctionBar_synthesizeEvent(bar, mevent.x);
            }

         }
      }

      switch (ch) {
      case ERR:
         refreshTimeout--;
         continue;
      case KEY_RESIZE:
         ListBox_resize(lb, COLS, LINES-headerHeight-1);
         if (incSearchMode)
            FunctionBar_draw(searchBar, incSearchBuffer);
         else
            FunctionBar_draw(defaultBar, NULL);
         break;
      case 'M':
      {
         refreshTimeout = 0;
         pf->sortKey = PERCENT_MEM;
         headerString = RichString_new();
         ProcessFilter_display((Object*)pf, &headerString);
         ListBox_setHeader(lb, headerString);
         break;
      }
      case 'P':
      {
         refreshTimeout = 0;
         pf->sortKey = PERCENT_CPU;
         headerString = RichString_new();
         ProcessFilter_display((Object*)pf, &headerString);
         ListBox_setHeader(lb, headerString);
         break;
      }
      case KEY_F(1):
      case 'h':
      {
         showHelp();
         FunctionBar_draw(defaultBar, NULL);
         refreshTimeout = 0;
         break;
      }
      case '\014': // Ctrl+L
      {
         clear();
         FunctionBar_draw(defaultBar, NULL);
         refreshTimeout = 0;
         break;
      }
      case KEY_ENTER:
      {
         doRefresh = true;
         refreshTimeout = 0;
         break;
      }
      case ' ':
      {
         Process* p = (Process*) ListBox_getSelected(lb);
         Process_toggleTag(p);
         ListBox_onKey(lb, KEY_DOWN);
         break;
      }
      case 'S':
      case KEY_F(2):
      {
         Setup_run(settings, headerHeight);
         headerHeight = Settings_calculateHeaderHeight(settings);
         ListBox_move(lb, 0, headerHeight);
         ListBox_resize(lb, COLS, LINES-headerHeight-1);
         FunctionBar_draw(defaultBar, NULL);
         break;
      }
      case 'F':
      {
         follow = true;
         continue;
      }
      case KEY_F(9):
      case 'k':
      {
         const int lbkWidth = 15;

         int sigCount = Signal_getSignalCount();
         if (!signals) {
            signals = Signal_getSignalTable();
            lbk = ListBox_new(0, headerHeight, lbkWidth-1, LINES - headerHeight - 2, SIGNAL_CLASS, true);
            for(int i = 0; i < sigCount; i++)
               ListBox_set(lbk, i, (Object*) signals[i]);
         }
         ListBox_setHeader(lbk, RichString_quickString("Send signal:"));
         ListBox_setSelected(lbk, 16); // 16th item is SIGTERM
         lbk->state.i = 0;
         lbk->eventHandler = SignalListBox_eventHandler;

         char* fuFunctions[2] = {"Send  ", "Cancel "};
         char* fuKeys[2] = {"Enter", "Esc"};
         int fuEvents[2] = {13, 27};
         FunctionBar* fuBar = FunctionBar_new(2, fuFunctions, fuKeys, fuEvents);
         
         ScreenManager* scr = ScreenManager_new(0, headerHeight, 0, -1, HORIZONTAL);
         ScreenManager_add(scr, lbk, lbkWidth - 1);
         ScreenManager_add(scr, lb, -1);
         ScreenManager_setFunctionBar(scr, fuBar);
         ListBox* lbFocus;
         int ch;
         ScreenManager_run(scr, &lbFocus, &ch);
         
         if (lbFocus == lbk && ch == 13) {
            Signal* signal = (Signal*) ListBox_getSelected(lbk);
            if (signal->number != 0) {
               ListBox_setHeader(lbk, RichString_quickString("Sending..."));
               ListBox_draw(lbk, true);
               refresh();
               bool anyTagged = false;
               for (int i = 0; i < ListBox_getSize(lb); i++) {
                  Process* p = (Process*) ListBox_get(lb, i);
                  if (p->tag) {
                     Process_sendSignal(p, signal->number);
                     Process_toggleTag(p);
                     anyTagged = true;
                  }
               }
               if (!anyTagged) {
                  Process* p = (Process*) ListBox_getSelected(lb);
                  Process_sendSignal(p, signal->number);
               }
               napms(500);
            }
         }
         
         FunctionBar_delete(fuBar);
         ScreenManager_delete(scr);
         
         ListBox_move(lb, 0, headerHeight);
         ListBox_resize(lb, COLS, LINES-headerHeight-1);
         FunctionBar_draw(defaultBar, NULL);

         break;
      }
      case KEY_F(10):
      case 'q':
         quit = 1;
         break;
      case '<':
      case ',':
      case KEY_F(5):
      {
         refreshTimeout = 0;
         ProcessFilter_sortKey(pf, -1);
         headerString = RichString_new();
         ProcessFilter_display((Object*)pf, &headerString);
         ListBox_setHeader(lb, headerString);
         break;
      }
      case '>':
      case '.':
      case KEY_F(6):
      {
         refreshTimeout = 0;
         ProcessFilter_sortKey(pf, 1);
         headerString = RichString_new();
         ProcessFilter_display((Object*)pf, &headerString);
         ListBox_setHeader(lb, headerString);
         break;
      }
      case 'I':
      case KEY_F(4):
      {
         refreshTimeout = 0;
         ProcessFilter_invertSortOrder(pf);
         break;
      }
      case KEY_F(8):
      case '[':
      {
         Process* p = (Process*) ListBox_getSelected(lb);;
         Process_setPriority(p, p->nice + 1);
         doRefresh = false;
         break;
      }
      case KEY_F(7):
      case ']':
      {
         Process* p = (Process*) ListBox_getSelected(lb);;
         Process_setPriority(p, p->nice - 1);
         doRefresh = false;
         break;
      }
      case KEY_F(3):
      case '/':
         FunctionBar_draw(searchBar, incSearchBuffer);
         incSearchMode = true;
         break;
      case 'C':
         showColumnConfig(pf);
         FunctionBar_draw(defaultBar, NULL);
         headerString = RichString_new();
         ProcessFilter_display((Object*)pf, &headerString);
         ListBox_setHeader(lb, headerString);
         refreshTimeout = 0;
         break;
      case 'T':
         pf->hideThreads = ~pf->hideThreads;
         break;
      default:
         doRefresh = false;
         refreshTimeout = resetRefreshTimeout;
         ListBox_onKey(lb, ch);
         break;
      }
      follow = false;
   }
   attron(COLOR_PAIR(WHITE_PAIR));
   mvhline(LINES-1, 0, ' ', COLS);
   attroff(COLOR_PAIR(WHITE_PAIR));
   refresh();
   
   CRT_done();
   Settings_write(settings);
   if (lbk) {
      ListBox_delete((Object*)lbk);
      // TODO: delete signals
   }
   TypedVector_delete(leftMeters);
   TypedVector_delete(rightMeters);
   ProcessList_delete(pl);
   FunctionBar_delete(searchBar);
   FunctionBar_delete(defaultBar);
   ListBox_delete((Object*)lb);
   UsersTable_delete(ut);
   ProcessFilter_delete(pf);
   Settings_delete(settings);
   if (signals)
      free(signals);
   debug_done();
   return 0;
}
